unit ucidetector;

{ http://www.tek-tips.com/faqs.cfm?fid=7402
  http://forum.lazarus.freepascal.org/index.php/topic,14956.msg128820.html#msg128820}

{$CODEPAGE UTF8}
{$MODE DELPHI}

interface

function IsUciEngine(const aExeName: string; out aName, aAuthor: string): boolean;

implementation

uses
  Windows, Classes, SysUtils, RegExpr;

var
  pci: TProcessInformation;
  ipr, ipw, opr, opw, epr, epw: THandle; { input/ouput/error pipe read/write }

function StartEngine(const aEngineName: string): boolean;
var
  si: TStartupInfo;
  sa: TSecurityAttributes;
begin
  with sa do
  begin
    nLength := SizeOf(TSecurityAttributes);
    bInheritHandle := TRUE;
    lpSecurityDescriptor := nil;
  end;
  
  CreatePipe(ipr, ipw, @sa, 0);
  CreatePipe(opr, opw, @sa, 0);
  CreatePipe(epr, epw, @sa, 0);
  
  FillChar(si, SizeOf(TStartupInfo), #0);
  
  with si do
  begin
    cb := SizeOf(TStartupInfo);
    hStdInput := ipr;
    hStdOutput := opw;
    hStdError := {epw}opw;
    dwFlags := STARTF_USESTDHANDLES or STARTF_USESHOWWINDOW;
    wShowWindow := SW_HIDE;
  end;
  
  result := CreateProcess(
    nil,
    pchar(aEngineName),
    @sa,
    @sa,
    TRUE,
    CREATE_NEW_CONSOLE or SYNCHRONIZE,
    nil,
    nil,
    si,
    pci
  );
end;

procedure WritePipeOut(aOutputPipe: THandle; s: string);
var
  w: DWORD = 0;
begin
  s := s + #13#10;
  WriteFile(aOutputPipe, s[1], Length(s), w, nil);
end;

function ReadPipeInput(aInputPipe: THandle): string;
var
  buffer: array[1..32767] of char;
  r: longword;
  sz: integer;
begin
  result := '';
  sz := SizeOf(buffer);
  PeekNamedPipe(aInputPipe, nil, sz, @r, @sz, nil);
  if r > 0 then
  begin
    ReadFile(aInputPipe, buffer, sz, r, nil);
    result := buffer;
    SetLength(result, r);
  end;
end;

procedure CloseHandles;
begin
  CloseHandle(pci.hProcess);
  CloseHandle(pci.hThread);
  CloseHandle(ipr);
  CloseHandle(ipw);
  CloseHandle(opr);
  CloseHandle(opw);
  CloseHandle(epr);
  CloseHandle(epw);
end;

function IsUciEngine(const aExeName: string; out aName, aAuthor: string): boolean;
const
  DELAY = 500;
var
  exeOutput: string;
  exitCode: DWORD;
begin
  result := FALSE;
  aName := '';
  aAuthor := '';
  if SetCurrentDir(ExtractFileDir(aExeName))
  and StartEngine(ExtractFileName(aExeName)) then
  begin
    Sleep(DELAY);    
    { Envoyons au processus la commande "uci" et attendons un peu. Pour bien faire il faudrait
      utiliser un *thread* et consulter périodiquement la réponse du processus jusqu'à ce que le
      mot-clé "uciok" soit apparu ou qu'un certain temps soit passé sans succès. }
    WritePipeOut(ipw, 'uci');
    Sleep(DELAY);   
    { Lisons la réponse du processus. }
    exeOutput := ReadPipeInput(opr);    
    { Sauvegardons dans un fichier texte la réponse du processus afin de pouvoir l'étudier. }
    with TStringList.Create do
    try
      Append(exeOutput);
      SaveToFile(ExtractFilePath(ParamStr(0)) + ChangeFileExt(ExtractFileName(aExeName), '.txt'));
    finally
      Free;
    end;    
    { Utilisons une expression régulière pour analyser la réponse du processus et, le cas échéant,
      en extraire les données qui nous intéressent, à savoir le nom du moteur et celui de son
      auteur. } 
    with TRegExpr.Create('id name ([^\r\n]+).+id author ([^\r\n]+).+uciok') do
    try
      if Exec(exeOutput) then
      begin
        result := TRUE;
        aName := Trim(Match[1]);
        aAuthor := Trim(Match[2]);
      end;
    finally
      Free;
    end;    
    { Si l'exécutable est bien un moteur UCI, demandons l'arrêt du processus en nous servant du
      protocole, plus précisément de la commande "quit". Dans le cas contraire, forçons l'arrêt du
      processus au moyen de la fonction TerminateProcess. }
    if result then
      WritePipeOut(ipw, 'quit')
    else
      TerminateProcess(pci.hProcess, 0);   
    { Attendons l'arrêt effectif du processus avant de libérer les *handles*. }
    GetExitCodeProcess(pci.hProcess, exitCode);
    while exitCode = STILL_ACTIVE do
    begin
      Sleep(DELAY); { Attente de l''arrêt du processus... }
      GetExitCodeProcess(pci.hProcess, exitCode);
    end;    
    CloseHandles;
  end;
end;

end.